/**
 * @file    shellmatta_transport.c
 * @brief   transport layer functions of shellmatta
 * @author  Simon Fischer <fischer.simon.1991@gmail.com>
 */

#include "shellmatta_transport.h"
#include "shellmatta.h"
#include "shellmatta_crc.h"
#include <string.h>



/**
 * @brief   create a secured packet and pass it to the shellmatta write function
 *
 * @note    The packetType, payloadLength and payload have to be preset!
 *
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @param[in]       packetType      type of packet to send
 * @param[in]       packet          packet to send
 * @return          errorcode
 */
static shellmatta_retCode_t shellmatta_transport_send(shellmatta_transport_layer_t *transportLayer,
                                                      shellmatta_transport_packet_t *packet)
{
    uint32_t checksum;
    char *rawPacketBuffer = (char*)packet;

    /** -# fill header */
    packet->startOfHeader   = SHELLMATTA_START_OF_HEADER;
    packet->protocolVersion = PROTOCOL_VERSION;
    packet->source          = 0;
    packet->destination     = 0;
    packet->sequenceH2S     = transportLayer->sequenceH2S;
    packet->sequenceS2H     = ++transportLayer->sequenceS2H;

    checksum = crc32Calc((char*) packet, SHELLMATTA_HEADER_LENGTH + packet->payloadLength);

    /* append crc to end of payload */
    rawPacketBuffer[SHELLMATTA_HEADER_LENGTH + packet->payloadLength] = (uint8_t)(checksum >> SHELLMATTA_SHIFT_BY_24);
    rawPacketBuffer[SHELLMATTA_HEADER_LENGTH + packet->payloadLength + SHELLMATTA_SHIFT_BY_1] = (uint8_t)(checksum >> SHELLMATTA_SHIFT_BY_16);
    rawPacketBuffer[SHELLMATTA_HEADER_LENGTH + packet->payloadLength + SHELLMATTA_SHIFT_BY_2] = (uint8_t)(checksum >> SHELLMATTA_SHIFT_BY_8);
    rawPacketBuffer[SHELLMATTA_HEADER_LENGTH + packet->payloadLength + SHELLMATTA_SHIFT_BY_3] = (uint8_t)(checksum);

    /* use original write function to send full buffer */
    return transportLayer->originalWrite((char*) rawPacketBuffer, SHELLMATTA_HEADER_LENGTH + packet->payloadLength + SHELLMATTA_LENGTH_CRC);
}


/**
 * @brief           Initializes the transportLayerInst
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @param[in]       writeFct        function pointer to output function
 * @return          errorcode       #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_transport_init( shellmatta_transport_layer_t    *transportLayer,
                                                shellmatta_write_t              writeFct)
{
    memset(transportLayer, 0u, sizeof(shellmatta_transport_layer_t));
    transportLayer->originalWrite = writeFct;

    return SHELLMATTA_OK;
}

/**
 * @brief           Resets all values of tranportLayerInst except for sequence counters
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @return          errorcode       #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_transport_reset(shellmatta_transport_layer_t *transportLayer)
{
    uint8_t sequenceH2S = transportLayer->sequenceH2S;
    uint8_t sequenceS2H = transportLayer->sequenceS2H;
    memset(transportLayer, 0u, sizeof(shellmatta_transport_layer_t));

    transportLayer->sequenceH2S = sequenceH2S;
    transportLayer->sequenceS2H = sequenceS2H;

    return SHELLMATTA_OK;
}

/**
 * @brief           processes the passed amount of data
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @param[in]       byte            byte to process
 * @param[out]      data            payload of processed packet - or manual mode
 * @param[out]      length          length of processed packet
 * @return          errorcode       #SHELLMATTA_OK      data is returned\n
 *                                  #SHELLMATTA_BUSY    packet reception ongoing\n
 *                                  #SHELLMATTA_ERROR   in case of crc error
 */
shellmatta_retCode_t shellmatta_transport_process(shellmatta_transport_layer_t   *transportLayer,
                                                  char                           byte,
                                                  char                           **data,
                                                  uint32_t                       *length)
{
    shellmatta_retCode_t ret = SHELLMATTA_BUSY;
    char *rawPacketBuffer = (char*)&transportLayer->packetBuffer;
    shellmatta_transport_packet_int_t intPacket;
    uint32_t refCrc;

    switch (transportLayer->state)
    {
    /* look for start of header */
    case SHELLMATTA_TRANSPORT_STATE_WAIT:
        /* if start of header is found, continue transport layer fsm */
        if (SHELLMATTA_START_OF_HEADER == byte)
        {
            memset(&transportLayer->packetBuffer, 0, sizeof(transportLayer->packetBuffer));
            transportLayer->headerIndex = 1u;
            transportLayer->payloadIndex = 0u;
            transportLayer->crcIndex = 0u;
            transportLayer->packetBuffer.startOfHeader = byte;
            transportLayer->state = SHELLMATTA_TRANSPORT_STATE_GET_HEADER;
        }
        else
        {
            /* just pass the payload as is */
            transportLayer->packetBuffer.payload[0] = byte;
            *data = transportLayer->packetBuffer.payload;
            *length = 1;
            transportLayer->active = false;
            ret = SHELLMATTA_OK;
        }
        break;

    case SHELLMATTA_TRANSPORT_STATE_GET_HEADER:

        rawPacketBuffer[transportLayer->headerIndex] = byte;
        transportLayer->headerIndex ++;

        if (transportLayer->headerIndex == SHELLMATTA_HEADER_LENGTH)
        {
            //TODO check for errors in header data
            if (0u != transportLayer->packetBuffer.payloadLength)
            {
                transportLayer->state = SHELLMATTA_TRANSPORT_STATE_GET_PAYLOAD;
            }
            else
            {
                transportLayer->state = SHELLMATTA_TRANSPORT_STATE_GET_CRC;
            }
        }

        break;

    /* read payload for previously read length of payload */
    case SHELLMATTA_TRANSPORT_STATE_GET_PAYLOAD:

        transportLayer->packetBuffer.payload[transportLayer->payloadIndex] = byte;
        transportLayer->payloadIndex ++;

        if (transportLayer->payloadIndex >= transportLayer->packetBuffer.payloadLength)
        {
            transportLayer->state = SHELLMATTA_TRANSPORT_STATE_GET_CRC;
        }
        break;

    /* read crc32 for four bytes */
    case SHELLMATTA_TRANSPORT_STATE_GET_CRC:

        transportLayer->crcIndex ++;
        transportLayer->packetBuffer.checksum |= ((uint8_t)byte << ((SHELLMATTA_LENGTH_CRC - transportLayer->crcIndex) * 8u));

        if (transportLayer->crcIndex >= SHELLMATTA_LENGTH_CRC)
        {
            refCrc = crc32Calc(rawPacketBuffer, SHELLMATTA_HEADER_LENGTH + transportLayer->packetBuffer.payloadLength);

            /* if crc is correct, further handling of data depends on type of packet */
            if (transportLayer->packetBuffer.checksum == refCrc)
            {

                transportLayer->active = true;

                /* crc is correct */
                transportLayer->sequenceH2S++;

                switch (transportLayer->packetBuffer.packetType)
                {
                case SHELLMATTA_TRANSPORT_PACKET_DATA:
                    /** return pointer to payload and length */
                    *data = (char*)&transportLayer->packetBuffer.payload;
                    *length = transportLayer->packetBuffer.payloadLength;
                    ret = SHELLMATTA_OK;
                    break;
                
                case SHELLMATTA_TRANSPORT_PACKET_SEQ_CNT_REQUEST:
                    /* send out packet with no payload */
                    intPacket.packetType = SHELLMATTA_TRANSPORT_PACKET_SEQ_CNT_RESPOND;
                    intPacket.payloadLength = 0u;
                    (void)shellmatta_transport_send(transportLayer, (shellmatta_transport_packet_t *)&intPacket);
                    break;

                case SHELLMATTA_TRANSPORT_PACKET_SEQ_CNT_RESPOND:
                    break;
                
                case SHELLMATTA_TRANSPORT_PACKET_MAX_BUFFERSIZE_REQUEST:
                    break;

                case SHELLMATTA_TRANSPORT_PACKET_MAX_BUFFERSIZE_RESPOND:
                    break;
                
                case SHELLMATTA_TRANSPORT_PACKET_SEARCH_DEVICE_REQUEST:
                    break;
                
                case SHELLMATTA_TRANSPORT_PACKET_SEARCH_DEVICE_RESPOND:
                    break;

                case SHELLMATTA_TRANSPORT_PACKET_SET_ADDRESS_REQUEST:
                    break;

                case SHELLMATTA_TRANSPORT_PACKET_SET_ADDRESS_RESPOND:
                    break;

                /* wrong packet type */
                default:
                    /* undo sequence counter increment */
                    transportLayer->sequenceH2S--;
                    break;
                }
            }
            else
            {
                /* crc is incorrect */
                ret = SHELLMATTA_ERROR;
            }

            /* reset state machine */
            transportLayer->state = SHELLMATTA_TRANSPORT_STATE_WAIT;
        }
        break;
    
    default:
        break;
    }

    return ret;
}

/**
 * @brief       Wrapper function for the write-function of shellmatta handle
 * 
 * This function is used to transmit data with the tranport layer protocol.\n
 * The input data is copied into a new array along with the header and crc32.\n
 * The resulting buffer is the forwarded to the original write function.\n
 * 
 * @note        If length of data exceeds #UINT8_MAX, data is sent in multiple packets.\n
 * 
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @param[in]       data            pointer to input data to process
 * @return          errorcode       #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_transport_write(shellmatta_transport_layer_t *transportLayer,
                                                const char* data,
                                                uint32_t length)
{
    shellmatta_retCode_t ret = SHELLMATTA_OK;
    uint32_t outPayloadLength = length;
    shellmatta_transport_packet_t *packet = &transportLayer->outputPacketBuffer;

    /* handle data with length bigger than max length */
    uint32_t processedLength    = 0u;
    uint32_t piledUpPayload;
    uint32_t splitLength;
    uint32_t splitPayloadLength;
    uint32_t restOfPayload;

    /* foot-controlled loop to send packets without payload length */
    do
    {
        piledUpPayload = transportLayer->outputPacketBuffer.payloadLength;

        /* compute length of next payload split */
        restOfPayload = (outPayloadLength - processedLength);
        if (restOfPayload > SHELLMATTA_TRANPORT_PAYLOAD_MAXLENGTH)
        {
            splitLength = SHELLMATTA_TRANPORT_PAYLOAD_MAXLENGTH;
        }
        else
        {
            splitLength = (restOfPayload + SHELLMATTA_HEADER_LENGTH + SHELLMATTA_LENGTH_CRC) % SHELLMATTA_TRANPORT_PAYLOAD_MAXLENGTH;
        }
        splitPayloadLength = splitLength - SHELLMATTA_HEADER_LENGTH - SHELLMATTA_LENGTH_CRC;

        (void)memcpy(&packet->payload[piledUpPayload], &data[processedLength], splitPayloadLength);
        packet->payloadLength += splitPayloadLength;
        processedLength += splitPayloadLength;

        if(packet->payloadLength >= SHELLMATTA_TRANPORT_PAYLOAD_MAXLENGTH)
        {
            /** -# packet is full - send */
            packet->packetType = SHELLMATTA_TRANSPORT_PACKET_DATA;
            ret = shellmatta_transport_send(transportLayer, &transportLayer->outputPacketBuffer);
            transportLayer->outputPacketBuffer.payloadLength = 0u;
        }
    }
    while (processedLength < outPayloadLength);
    return ret;
}

/**
 * @brief           Send out piled up payload
 * 
 * @param[in, out]  transportLayer  transport layer instance to work on
 * @return          errorcode       #SHELLMATTA_OK
 */
shellmatta_retCode_t shellmatta_transport_flush(shellmatta_transport_layer_t *transportLayer)
{
    shellmatta_retCode_t ret = SHELLMATTA_OK;
    shellmatta_transport_packet_t *packet = &transportLayer->outputPacketBuffer;

    if(0 != transportLayer->outputPacketBuffer.payloadLength)
    {
        packet->packetType = SHELLMATTA_TRANSPORT_PACKET_DATA;
        ret = shellmatta_transport_send(transportLayer, &transportLayer->outputPacketBuffer);
        packet->payloadLength = 0u;
    }

    return ret;
}
